from theia.cli.parser import get_parent_parser
from theia.cli.watcher import get_parser as get_watcher_parser, run_watcher
from theia.cli.collector import get_parser as get_collector_parser, run_collector
from theia.cli.query import get_parser as get_query_parser, run_query

parser = get_parent_parser('theia', 'Theia CLI')

subparsers = parser.add_subparsers(dest='command', title='command', help='CLI commands')
get_watcher_parser(subparsers)
get_collector_parser(subparsers)
get_query_parser(subparsers)

args = parser.parse_args()

if args.version:
    from theia.metadata import version
    from sys import exit
    print('theia', version)
    exit(0)

if args.verbose:
    from logging import basicConfig, DEBUG
    basicConfig(level=DEBUG)


if args.command == 'watch':
    run_watcher(args)
elif args.command == 'collect':
    run_collector(args)
elif args.command == 'query':
    run_query(args)



"""
-----------
theia.query
-----------

Query theia server for events.

This module contains the :class:`Query` that implements API for querying theia collector server.
"""
import asyncio
import json
from logging import getLogger
from theia.comm import Client


log = getLogger(__name__)


class ResultHandler:
    """Represents a result of a query against a theia collector server.

    The result of a query to the collector server is a stream of events. A callback (see ``callback`` in
    :meth:`Query.live`) can be set to be called whenever an :class:`theia.model.Event` is received, but the control to
    close the stream is passed down to a :class:`ResultHandler`.

    This class wraps the :class:`theia.comm.Client` used to connect to the collector server and adds support for
    registering on client closed events. It also adds means to close (cancel) the connection to the theia server.

    :param client: :class:`theia.comm.Client`, the underlying client to the theia server.

    """
    def __init__(self, client):
        self.client = client
        self._close_handlers = []
        self.client.on_close(self._on_client_closed)

    def _on_client_closed(self, websocket, code, reason):
        for hnd in self._close_handlers:
            try:
                hnd(self.client, code, reason)
            except Exception as e:
                log.debug('ResultHandler[%s]: close hander %s error: %s', self.client, hnd, e)

    def when_closed(self, closed_handler):
        """Register a handler to be called when the connection to the server is closed.

        The handler has the following signature:

        .. code-block:: python

            def closed_handler(client, code, reason):
                pass

        where:

        * ``client`` - :class:`theia.comm.Client`, is the underlying client connected to the theia collector.
        * ``code`` - ``int``, wbsocket close connection code.
        * ``reason`` - ``str``, the reason for closing the connection.

        """
        self._close_handlers.append(closed_handler)

    def cancel(self):
        """Cancel this result.

        Closes the underlying client connection.
        """
        if self.client.is_open():
            self.client.close()


class Query:
    """Represents a query to a theia collector server.

    The query instances are thread-safe.

    :param host: ``str``, the hostname (IP) of the collector server.
    :param port: ``int``, the port of the collector server.
    :param secure: ``bool``, whether to connect securely to the collector server.
    :param loop: :class:`asyncio.BaseEventLoop`, the event loop to use.

    """
    def __init__(self, host, port, secure=False, loop=None):
        self.connections = set()
        self.host = host
        self.port = port
        self.secure = secure
        self.loop = loop if loop is not None else asyncio.get_event_loop()

    def live(self, criteria, callback=None):
        """Make a live query to the collector.

        The collector will watch for any events that occur **after** the live query is registered and return those that
        match the given ``criteria``.

        :param criteria: ``dict``, the criteria filter for the events. This is a ``dict`` that can contain the following
            keys:

            * ``id``, ``str``, pattern for matching the event id.
            * ``source``, ``str``, pattern for matching the event source.
            * ``start``, ``int``, match events with timestamp greater or equal to this.
            * ``end``, ``int``, match events with timestamp less than or equal to this.
            * ``content``, ``str``, pattern for matching the event content.
            * ``tags``, ``list``, list of patterns to match against the event tags.

        :param callback: ``function``, called with the matching event content. The method takes one argument, the
            serialized :class:`theia.model.Event`.

        Returns a :class:`ResultHandler`.
        """
        return self._connect_and_send('/live', criteria, callback)

    def find(self, criteria, callback=None):
        """Make a query to the collector.

        The collector will search for any events that occured **before** theis query was sent to the server and will
        return those that match the given ``criteria``.

        :param criteria: ``dict``, the criteria filter for the events. This is a ``dict`` that can contain the following
            keys:

            * ``id``, ``str``, pattern for matching the event id.
            * ``source``, ``str``, pattern for matching the event source.
            * ``start``, ``int``, match events with timestamp greater or equal to this.
            * ``end``, ``int``, match events with timestamp less than or equal to this.
            * ``content``, ``str``, pattern for matching the event content.
            * ``tags``, ``list``, list of patterns to match against the event tags.

        :param callback: ``function``, called with the matching event content. The method takes one argument, the
            serialized :class:`theia.model.Event`.

        Returns a :class:`ResultHandler`.
        """
        return self._connect_and_send('/find', criteria, callback)

    def _connect_and_send(self, path, criteria, callback):

        client = Client(loop=self.loop, host=self.host, port=self.port, path=path, recv=callback)

        client.connect()

        self.connections.add(client)

        def on_client_closed(websocket, code, reason):
            """Remove the underlying client from the connections set once its connection is closed.
            """
            # client was closed
            if client in self.connections:
                self.connections.remove(client)

        client.on_close(on_client_closed)

        msg = json.dumps(criteria)

        client.send(msg)

        return ResultHandler(client)



"""
----------
theia.comm
----------

Theia communication module.

Defines classes for building theia servers and clients.

Theia communication module is asynchronous and is built on top of asyncio loop.

There are two main interfaces:

* ``Server`` an async server handling and managing WebSocket connections from multiple clients.
* ``Client`` an async client connection to a theia server.

The interfaces are designed to work primarily with theia events and are thread-safe.
"""

import asyncio
import json
from logging import getLogger
import websockets
from theia.model import EventSerializer


log = getLogger(__name__)


_WAIT_TIME = 0.1


class Client:
    """Client represents a client connection to a theia server.

    :param loop: :mod:`asyncio` EventLoop to use for this client.
    :param host: ``str``, theia server hostname.
    :param port: ``int``, theia server port.
    :param secure: ``bool``, is the connection secure.
    :param path: ``str``, the request path - for example: ``"/live"``, ``"/events"`` etc.
    :param recv: ``function``, receive handler. Called when a message is received from the
        server. The handler has the following signature:

        .. code-block:: python

            def handler(message):
                pass

    where:
        * ``message`` is the message received from the theia server.
    """

    def __init__(self, loop, host, port, secure=False, path=None, recv=None):
        self.loop = loop
        self.host = host
        self.port = port
        self.secure = secure
        self.path = path
        self.recv_handler = recv
        self.serializer = EventSerializer()
        self.websocket = None
        self._is_open = False
        self._close_handlers = []

    async def _open_websocket(self):
        websocket = await websockets.connect(self._get_ws_url(), loop=self.loop)
        self.websocket = websocket
        self._is_open = True
        asyncio.ensure_future(self._recv(), loop=self.loop)
        log.debug('[%s:%d]: connected', self.host, self.port)

    def connect(self):
        """Connect to the remote server.
        """
        self.loop.run_until_complete(self._open_websocket())

    def close(self, reason=None):
        """Close the connection to the remote server.

        :param reason: ``str``, the reason for disconnecting. If not given, a default ``"normal close"`` is
            sent to the server.

        """
        reason = reason or 'normal close'
        self._is_open = False
        self.websocket.close(code=1000, reason=reason)
        log.debug('[%s:%d]: explicitly closed. Reason=%s', self.host, self.port, reason)

    def _get_ws_url(self):
        url = 'wss://' if self.secure else 'ws://'
        url += self.host
        if self.port:
            url += ':' + str(self.port)
        if self.path:
            if self.path.startswith('/'):
                url += self.path
            else:
                url += '/' + self.path
        return url

    def send(self, message):
        """Send a ``str`` message to the remote server.

        :param message: ``str``, the message to be sent to the remote server.

        Returns the :class:`asyncio.Handle` to the scheduled task for sending the
        actual data.
        """
        return self.loop.call_soon_threadsafe(self._call_send, message)

    def _call_send(self, message):
        asyncio.ensure_future(self.websocket.send(message), loop=self.loop)

    def send_event(self, event):
        """Send an event to the remote server.

        Serializes, then sends the serialized content to the remote server.

        :param event: :class:`theia.model.Event`, the event to be send.

        Returns the :class:`asyncio.Handle` to the scheduled task for sending the
        actual data.
        """
        message = self.serializer.serialize(event)
        return self.send(message)

    async def _recv(self):
        while self._is_open:
            try:
                message = await self.websocket.recv()
                await self._process_message(message)
            except websockets.ConnectionClosed as wse:
                self._closed(wse.code, wse.reason)
                log.debug('[%s:%d] connection closed', self.host, self.port)
            # pylint: disable=broad-except
            # General case
            except Exception as e:
                log.exception(e)
                self._closed(1006, reason=str(e))

    async def _process_message(self, message):
        if self.recv_handler:
            self.recv_handler(message)

    def on_close(self, handler):
        """Add close handler.

        The handles is called when the client connection is closed either by the client
        or by the server.

        :param handler: ``function``, the handler callback. The callback prototype
            looks like so:

            .. code-block:: python

                def callback(websocket, code, reason):
                    pass

        where:

        * ``websocket`` :class:`websockets.WebSocketClientProtocol` is the underlying
            websocket.
        * ``code`` ``int`` is the code received when the connection was closed. Check
            out the `WebSocket specification`_ for the list of codes and their meaning.
        * ``reason`` ``str`` is the reason for closing the connection.

        .. _WebSocket specification: https://tools.ietf.org/html/rfc6455#section-7.4
        """
        self._close_handlers.append(handler)

    def _closed(self, code=1000, reason=None):
        self._is_open = False
        for hnd in self._close_handlers:
            try:
                hnd(self.websocket, code, reason)
            except Exception as e:
                log.debug(e)

    def is_open(self):
        """Check if the client connection is open.

        Returns ``True`` if the client connection is open, otherwise ``False``.
        """
        return self._is_open


class wsHandler:
    """Wrapper for an incoming websocket connection.

    Used primarily with the :class:`Server` implementation in the client connections
    life-cycle management.

    :param websocket: :class:`websockets.WebSocketClientProtocol`, the underlying
        websocket connection.
    :param path: ``str``, the request path of the websocket connection.

    **Note**: This class is mainly used internally and as such it is a subject of
    chnages in its API.
    """
    def __init__(self, websocket, path):
        self.ws = websocket
        self.path = path
        self.close_handlers = []

    def trigger(self, websocket):
        """Triggers the close handlers for this websocket.

        :param websocket: :class:`websockets.WebSocketClientProtocol`, the underlying
            websocket connection.
        """
        for hnd in self.close_handlers:
            try:
                hnd(self.ws, self.path)
            # pylint: disable=broad-except
            except Exception as ex:
                log.debug(ex)

    def add_close_handler(self, hnd):
        """Register a close handler for this connection.

        :param hnd: ``function``, the close handler callback. The callback receives
            two parameters:

            * ``ws`` (:class:`websockets.WebSocketClientProtocol`), the underlying websocket
                connection.
            * ``path`` (``str``), the request path of the websocket.
        """
        self.close_handlers.append(hnd)


class Server:
    """Listens for and manages multiple client connections.

    The server is based on :mod:`asyncio` event loop. It manages the websocket
    connections comming from multiple clients based on the path in the websocket
    request connection.

    Provides a way to register a callback for notifying when a client connects to
    a particular endpoint (path), and also a way to register a callback for when
    the client disconnects.

    Instances of this class are thread-safe.

    :param loop: :class:`asyncio.BaseEventLoop`, the event loop.
    :param host: ``str``, the hostname to bind to when listening fo incoming
        connections.
    :param port: ``int``, the port to listen on.
    """
    def __init__(self, loop, host='localhost', port=4479):
        self.loop = loop
        self.host = host
        self.port = port
        self.websockets = {}
        self._started = False
        self.actions = {}
        self._stop_timeout = 10

    def on_action(self, path, cb):
        """Register a callback to listen for messages from clients that connected
        to this specific entrypoint (path).

        The callback will be called whenever a new message is received from the client
        on this ``path``.

        If multiple callbacks are registered on the same action, then they are called
        one by one in the same order as registered. The response from the callbacks is
        chained between the subsequent calls.

        :param path: ``str``, the request path of the incoming websocket connection.
        :param cb: ``function``, the callback to be called when a message is received
            from the client on this path.
            The callback handler looks like this:

            .. code-block:: python

                def callback(path, message, websocket, resp):
                    return resp

        where:

        * ``path`` ``str``, the path on which the message was received.
        * ``message`` ``str``, the messge received from the websocket connection.
        * ``websocket`` :class:`websockets.WebSocketClientProtocol`, the underlying
            websocket.
        * ``resp`` ``str``, the response from the previous action registered on this
            same action

        The callback must return ``str`` response or ``None``.
        """
        actions = self.actions.get(path)
        if not actions:
            actions = self.actions[path] = []
        actions.append(cb)

    async def _on_client_connection(self, websocket, path):
        self.websockets[websocket] = wsHandler(websocket, path)
        try:
            while self._started:
                message = await websocket.recv()
                resp = await self._process_req(path, message, websocket)
                if resp is not None:
                    await websocket.send(str(resp))
        except websockets.ConnectionClosed:
            self._remove_websocket(websocket)
            log.debug('[Server:%s:%d] Closing websocket connection: %s', self.host, self.port, websocket)
        except Exception as e:
            self._remove_websocket(websocket)
            log.error('[Server:%s:%d] Closing websocket connection because of unknown error: %s',
                      self.host, self.port, websocket)
            log.exception(e)

    def _remove_websocket(self, websocket):
        # self.websockets.remove(websocket)
        hnd = self.websockets.get(websocket)
        if hnd is not None:
            del self.websockets[websocket]
            hnd.trigger(websocket)

    def on_websocket_close(self, websocket, cb):
        """Register a close callback for this websocket.

        :param websocket: :class:`websockets.WebSocketClientProtocol`, the websocket
            to watch for closing.
        :param cb: ``function``, the callback to be called when the ``websocket``
            is closed. The callback should look like this:

            .. code-block:: python

                def callback(ws, path):
                    pass

        where:

        * ``ws`` (:class:`websockets.WebSocketClientProtocol`), the underlying websocket
                connection.
        * ``path`` (``str``), the request path of the websocket.

        This method returns ``True`` if the callback was added; ``False`` if the
        websocket is not managed by this :class:`Server` instance.
        """
        hnd = self.websockets.get(websocket)
        if hnd is not None:
            hnd.add_close_handler(cb)
            return True
        return False

    async def _process_req(self, path, message, websocket):
        resp = ''
        for reg_path, actions in self.actions.items():
            if reg_path == path:
                try:
                    for action in actions:
                        resp = action(path, message, websocket, resp)
                # pylint: disable=broad-except
                # Intended to be broad as it handles generic action
                except Exception as e:
                    log.exception(e)
                    return json.dumps({"error": str(e)})
                break
        return resp

    def start(self):
        """Starts the server.

        This call blocks until the server is started or an error occurs.
        """
        start_server = websockets.serve(self._on_client_connection,
                                        self.host, self.port, loop=self.loop)
        self.loop.run_until_complete(start_server)
        self._started = True

    def stop(self):
        """Stops the server.

        Closes all client websocket connections then shuts down the server.

        This operation blocks until the server stops or an error occurs.
        """
        from time import sleep
        wss = [ws for ws, _ in self.websockets.items()]
        semaphore = {'value': len(wss)}

        for ws in wss:
            try:
                self._remove_websocket(ws)
            except Exception as exc:
                log.exception(exc)
            asyncio.run_coroutine_threadsafe(self._send_close(semaphore, ws, 'server stop'), self.loop)
        log.debug('Server notified to stop')
        total_wait = 0
        while semaphore['value']:
            sleep(_WAIT_TIME)
            total_wait += _WAIT_TIME
            if total_wait > self._stop_timeout:
                break
        if semaphore['value']:
            log.warning('Stop timeout reached before all connections were closed. Server will stop anyway')
        else:
            log.debug('All done. Server stopped.')

    async def _send_close(self, semaphore, websocket, reason=None):
        await websocket.close(code=1000, reason=reason)
        semaphore['value'] -= 1



"""
----------------
theia.naivestore
----------------

Naive implementation of the Event Store.

This module provides an implementation of the :class:`theia.storeapi.EventStore`
that stores the events in plain-text files.

The store writes to the files atomically, so there is no danger of leaving the
files in an inconsistent state.

The files in which the store keeps the events are plain text files that contain
serialized events, encoded in UTF-8. The events are written sequentially. Plain
text is chosen so that these files can be also be read and processed by other
tools (such as grep). The events are kept in multiple files. Each file contains
about a minute worth of events - all events that happened in that one minute time
span. The name of the file is the time span: <first-event-timestamp>-<last-event-timestamp>.

The naive store requires a root directory in which to store the events. Here is
an example of usage of the store:

.. code-block:: python

    from theia.naivestore import NaiveEventStore
    from theia.model import Event
    from uuid import uuid4
    from datetime import datetime

    store = NaiveEventStore(root_dir='./data')

    timestamp = datetime.now().timestamp()

    store.save(Event(id=uuid4(),
                     source='test-example',
                     timestamp=timestamp,
                     tags=['example'],
                     content='event 1'))
    store.save(Event(id=uuid4(),
                     source='test-example',
                     timestamp=timestamp + 10,
                     tags=['example'],
                     content='event 2'))
    store.save(Event(id=uuid4(),
                     source='test-example',
                     timestamp=timestamp + 20,
                     tags=['example'],
                     content='event 3'))

    # now let's search some events

    for ev in store.search(ts_start=timestamp + 5):
        print('Found:', ev.content)

would print::

    >> Found: event 2
    >> Found: event 3

"""

from tempfile import NamedTemporaryFile
from io import BytesIO, SEEK_CUR
from threading import RLock, Thread
from shutil import move
from os.path import join as join_paths, basename, dirname
from os import listdir
from collections import namedtuple
import re
import time
from logging import getLogger
from theia.storeapi import EventStore
from theia.model import EventSerializer, EventParser, EOFException


log = getLogger(__name__)


class PeriodicTimer(Thread):
    """Timer that executes an action periodically with a given interval.

    The timer executes the action in a separate thread (as this class is a subclass of :class:`threading.Thread`). To
    run the action you must call :meth:`PeriodicTimer.start`.
    The first execution of the action is delayed by ``interval`` seconds. This timer does not call the ``action``
    callback every ``interval`` seconds, but rather waits ``interval`` seconds between after the action completes until
    the next call. So for a long running tasks, the time of call of the action may not be evenly spaced.

    You can cancel this timer by calling meth:`PeriodicTimer.cancel`.

    :param interval: ``numeric``, seconds to wait between subsequent calls to ``action`` callback.
    :param action: ``function``, the action callback. This callback takes no arguments.
    """
    def __init__(self, interval, action):
        super(PeriodicTimer, self).__init__(name='periodic-timer@%f:[%s]' % (interval, str(action)))
        self.interval = interval
        self.action = action
        self.is_running = False

    def run(self):
        """Runs the periodic timer.

        Do **not** call this function directly, but rather call :meth:`PeriodicTimer.start` to run this thread.

        To cancel the timer, call :meth:`PeriodicTimer.cancel`.
        """
        self.is_running = True

        time.sleep(self.interval)
        while self.is_running:
            try:
                self.action()
            except:
                pass
            time.sleep(self.interval)

    def cancel(self):
        """Cancels the running timer.

        The timer thread may continue running until the next cycle, then it exits.
        """
        self.is_running = False


class SequentialEventReader:
    """Reads events (:class:`theia.model.Event`) from an incoming :class:`io.BytesIO` stream.

    Uses an :class:`theia.model.EventParser` to parse the events from the incoming stream.

    Provides two ways of parsing the events:

    * Parsing the event fully - loads the header and the content of the event. See :meth:`SequentialEventReader.events`.
    * Parsing only the event header - this skips the loading of the content. Useful for not wasting performance/memory
        on loading and decoding the event content when not searching by the event content.

    This reader implements the context manager interface and can be used in ``with`` statements. For example:

    .. code-block:: python

        with SequentialEventReader(stream, parser) as reader:
            for event in reader.events():
                print(event)


    :param stream: :class:`io.BytesIO`, the incoming stream to read events from.
    :param event_parser: :class:`theia.model.EventParser`, the parser used for parsing the events from the stream.

    """
    def __init__(self, stream, event_parser):
        self.stream = stream
        self.parser = event_parser

    def events(self):
        """Reads full events from the incoming stream.

        Returns an iterator for the read events and yields :class:`theia.model.Event` as it becomes available in the
        stream.

        The iterator stops if there are no more events available in the stream or the stream closes.
        """
        while True:
            try:
                yield self._actual_read()
            except EOFException:
                break

    def events_no_content(self):
        """Reads events without content (just header) from the incoming stream.

        Returns an iterator for the read events and yields :class:`theia.model.Event` as it becomes available in the
        stream.

        Note that the ``content`` property of the :class:`theia.model.Event` will always be set to None.

        The iterator stops if there are no more events available in the stream or the stream closes.
        """
        while True:
            try:
                yield self._actual_read(skip_content=True)
            except EOFException:
                break

    def curr_event(self):
        """Reads an :class:`theia.model.Event` at the current position in the stream.

        Reads the event fully.

        Returns the :class:`theia.model.Event` at the current position of the stream or ``None`` if there are no more
        available events to be read from the stream (the stream closes).
        """
        try:
            return self._actual_read()
        except EOFException:
            return None

    def _actual_read(self, skip_content=False):
        data = self.parser.parse_event(self.stream, skip_content=skip_content)
        try:
            self.stream.seek(1, SEEK_CUR)
        except:
            pass
        return data

    def __enter__(self):
        """Implements the context-manager enter method.
        Returns a reference to itself.
        """
        return self

    def __exit__(self, *args):
        """Implements the context-manager exiting method.
        Closes the underlying stream.
        """
        self.stream.close()


class MemoryFile:
    """File-system backed in-memory buffer.

    This class wraps an :class:`io.BytesIO` buffer and backs it up with a real file in the file-system.
    The writes go to the in-memory buffer, which then can be flushed to the actual file in the file-system.

    The flushing of the buffer is atomic and consistent. The buffer is first flushed to a temporary file, then the
    system buffers are synced, and then the temporary file is renamed as the actual file.

    The instances of this class are thread-safe and can be shared between threads.

    **Limitations:** This class is not optimized for large data files as it keeps all of the file data in memory. This
    may cause a performance penalty in both speed and memory consumption when dealing with large files. In those cases
    it is better to use other memory mapped file implementations.

    :param name: ``str``, the name of the file in the file-system. This is just the filename, without the directory.
    :param path: ``str``, the directory holding the file in the file-system.
    """
    def __init__(self, name, path):
        self.name = name
        self.path = path
        self.buffer = BytesIO()
        self.lock = RLock()

    def write(self, data):
        """Write to the data to the buffer.

        This writes the data to the in-memory buffer.

        :param data: ``bytes``, the data to be written to the buffer.
        """
        try:
            self.lock.acquire()
            self.buffer.write(data)
        finally:
            self.lock.release()

    def stream(self):
        """Returns a copy of the underlying :class:`io.BytesIO` in-memory stream.
        """
        # copy the buffer
        # return the copy
        try:
            self.lock.acquire()
            return BytesIO(self.buffer.getvalue())
        finally:
            self.lock.release()

    def flush(self):
        """Writes the in-memory buffer to the file in the file-system.

        This operation is atomic and guarantees that the complete state of the buffer will be written to the file.
        The underlying file will never be left in an inconsistent state. This is achieved by first writing the entire
        buffer to a temporary file (in the same directory), flushing the system buffers, then if this succeeds, renaming
        the temporary file as the original file name.
        """
        tmpf = NamedTemporaryFile(dir=self.path, delete=False)
        try:
            self.lock.acquire()
            tmpf.write(self.buffer.getvalue())
            tmpf.flush()
            move(tmpf.name, join_paths(self.path, self.name))
        finally:
            self.lock.release()


DataFile = namedtuple('DataFile', ['path', 'start', 'end'])
"""Represents a file containing data (events) within a given time interval (from ``start`` to ``end``).
"""

DataFile.path.__doc__ = """
    ``str``, the path to the data file.
"""

DataFile.start.__doc__ = """
    ``int``, timestamp, the start of the interval. The data file does not contain any events before this timestamp.
"""

DataFile.end.__doc__ = """
    ``int``, timestamp, the end of the interval. The data file does not contain any events after this timestamp.
"""


def binary_search(datafiles, timestamp):
    """Performs a binary search in the list of datafiles, for the index of the first data file that contain events that
    happened at or after the provided timestamp.

    :param datafiles: ``list`` of :class:`DataFile`, list of datafiles **sorted** in ascending order by the start
        timestamp.
    :param timestamp: ``float``, the timestamp to serch for.

    Returns the index (``int``) of the first :class:`DataFile` in the list that contains event that occurred at or
    after the provided timestamp. Returns ``None`` if no such data file can be found.
    """
    start = 0
    end = len(datafiles) - 1
    if not datafiles:
        return None
    if datafiles[0].start > timestamp or datafiles[-1].end < timestamp:
        return None

    while True:
        mid = (end + start) // 2
        if datafiles[mid].end >= timestamp:
            end = mid
        else:
            start = mid
        if end - start <= 1:
            if datafiles[start].end >= timestamp:
                return start
            return end
    return None


class FileIndex:
    """An index of :class:`DataFile` files loaded from the given directory.

    Loads an builds an index of :class:`DataFile` files from the given directory. Each data file name must be in the
    form: ``<start>-<end>``, where ``start`` and ``end`` represent the time interval of the events in that data file.

    :class:`FileIndex` loads all data files and builds a total time span of all data files. The index exposes methods
    for locating and searching files that contain the events within a given time interval.

    :param root_dir: ``str``, the directory from which to load the data files.
    """
    def __init__(self, root_dir):
        self.root = root_dir
        self.files = self._load_files(root_dir)

    def _load_files(self, root_dir):
        files = []

        for file_name in listdir(root_dir):
            data_file = self._load_data_file(file_name)
            if data_file:
                files.append(data_file)

        files = sorted(files, key=lambda n: n.start)
        log.info('Loaded %d files to index.', len(files))
        if files:
            log.info('Spanning from %d to %d', files[0].start, files[-1].end)
        return files

    def _load_data_file(self, fname):
        if re.match(r'\d+-\d+', fname):
            start, _, end = fname.partition('-')
            start, end = int(start), int(end)
            return DataFile(path=join_paths(self.root, fname), start=start, end=end)
        return None

    def find(self, ts_from, ts_to):
        """Finds the data files that contain the events within the given interval [``ts_from``, ``ts_to``].

        The interval can be open at the end ([``ts_from``, ``Inf``]), by passing ``0`` for ``ts_to``. The interval
        cannot be open at the start.

        :param ts_from: ``float``, timestamp, find all data files containing events that have timestamp greater than or
            equal to ``ts_from``.
        :param ts_to: ``float``, timestamp, find all data files containing events that have timestamp less than or equal
            to ``ts_to``. If ``0`` or ``None`` is passed for this parameter, then the end of the time span is open,
            meaning this parameter will be ignore in the search.

        Returns a ``list`` of :class:`DataFile` files that contain the events within the given interval. Returns
        ``None`` if there are no files containing events within the given interval.
        """
        idx = binary_search(self.files, ts_from)
        if idx is None and self.files:
            if self.files[0].start >= ts_from:
                idx = 0
            elif self.files[-1].end <= ts_from:
                idx = len(self.files) - 1
        if idx is not None:
            found = []
            while idx < len(self.files):
                data_file = self.files[idx]
                if ts_to:
                    if data_file.start > ts_to:
                        break # hit the last
                if data_file.start >= ts_from or data_file.end >= ts_from:
                    found.append(data_file)

                idx += 1
            return found if found else None
        return None

    def find_event_file(self, timestamp):
        """Find the event file that contains the event with the given timestamp.

        :param timestamp: ``int``, the timestamp of the event.

        Returns the :class:`DataFile` containing the event, or ``None`` if it cannot be found.
        """
        idx = binary_search(self.files, timestamp)
        if idx is not None:
            return self.files[idx]
        return None

    def add_file(self, fname):
        """Add a data file to the :class:`FileIndex`.

        The time span will be recalculated to incorporate this new data file.

        :param fname: ``str``, the file name of the data file to be added to the file index.
        """
        data_file = self._load_data_file(fname)
        if data_file:
            self.files.append(data_file)
            self.files = sorted(self.files, key=lambda n: n.start)


class NaiveEventStore(EventStore):
    """A naive implementation of the :class:`theia.storeapi.EventStore` that keeps the event data is a plain text files.

    The events are kept in plain text files, serialized by default in UTF-8. The files are human readable and the format
    is designed to be (relatively) easily processed by other tools as well (such as ``grep``). Each data file contains
    events that happened within one minute (60000ms). The names of the data files also reflect the time span interval,
    so for example a file with name *1528631988-1528632048* contains only events that happened at or after
    ``1528631988``, but before ``1528632048``.

    The store by default uses in-memory buffers to write new events, and flushes the buffer periodically. By default
    the flushing occurs roughly every second (1000ms, see the parameter ``flush_interval``). This increases the
    performance of the store, but if outage occurs within this interval, the data in the in-memory buffers will be lost.
    The store can be configured to flush the events immediately on disk (by passing ``0`` for ``flush_interval``), but
    this decreases the performance of the store significantly.

    The instances of this class are thread-safe and can be shared between threads.

    :param root_dir: ``str``, the root directory where to store the events data files.
    :param flush_interval: ``int``, flush interval for the data files buffers in milliseconds. The event data files will
        be flushed and persisted on disk every ``flush_interval`` milliseconds. The default value is 1000ms. To flush
        immediately (no buffering), set this value equal or less than ``0``.
    """
    def __init__(self, root_dir, flush_interval=1000):
        self.root_dir = root_dir
        self.data_file_interval = 60  # 60 seconds
        self.serializer = EventSerializer()
        self.index = FileIndex(root_dir)
        self.open_files = {}
        self.write_lock = RLock()
        self.flush_interval = flush_interval  # <=0 immediate, otherwise will flush periodically
        self.timer = None
        if flush_interval > 0:
            self.timer = PeriodicTimer(flush_interval / 1000, self._flush_open_files)
            self.timer.start()
            log.info('Flushing buffers every %fms', (flush_interval / 1000))

    def _get_event_file(self, ts_from):
        data_file = self.index.find_event_file(ts_from)
        if not data_file:
            try:
                self.write_lock.acquire()
                data_file = self._get_new_data_file(ts_from)
                self.index.add_file(basename(data_file.path))
            finally:
                self.write_lock.release()
        return data_file

    def _open_file(self, data_file):
        return MemoryFile(name=basename(data_file.path), path=dirname(data_file.path))

    def _flush_open_files(self):
        for file_name, open_file in self.open_files.items():
            try:
                open_file.flush()
            except Exception as e:
                log.error('Error while flushing %s. Error: %s', file_name, e)

    def _get_new_data_file(self, ts_from):
        ts_end = ts_from + self.data_file_interval
        return DataFile(join_paths(self.root_dir, '%d-%d' % (ts_from, ts_end)), ts_from, ts_end)

    def save(self, event):
        # lookup file/create file
        # lock it
        # save serialized event
        # unlock
        data_file = self._get_event_file(event.timestamp)
        try:
            self.write_lock.acquire()
            if not self.open_files.get(data_file.path):
                self.open_files[data_file.path] = self._open_file(data_file)
            mem_file = self.open_files[data_file.path]
            mem_file.write(self.serializer.serialize(event))
            mem_file.write('\n'.encode(self.serializer.encoding))
            if self.flush_interval <= 0:
                mem_file.flush()
        finally:
            self.write_lock.release()

    def search(self, ts_start, ts_end=None, flags=None, match=None, order='asc'):
        data_files = self.index.find(ts_start, ts_end)
        if data_files:
            for data_file in data_files:
                yield from self._search_data_file(data_file, ts_start, ts_end, flags, match, order == 'desc')

    def close(self):
        self._flush_open_files()
        if self.timer:
            self.timer.cancel()
            self.timer.join()
        log.info('Naive Store stopped')

    def _search_data_file(self, data_file, ts_start, ts_end, flags, match, reverse):
        if reverse:
            yield from self._match_reverse(data_file, ts_start, ts_end, flags, match)
        else:
            yield from self._match_forward(data_file, ts_start, ts_end, flags, match)

    def _match_forward(self, data_file, ts_start, ts_end, flags, match):
        with self._seq_event_parser(data_file) as sqp:
            for event in sqp.events():
                if event.timestamp >= ts_start and (ts_end is None or event.timestamp <= ts_end):
                    if self._matches(event, flags, match):
                        yield event

    def _match_reverse(self, data_file, ts_start, ts_end, flags, match):
        matched = []
        for result in self._match_forward(data_file, ts_start, ts_end, flags, match):
            matched.append(result)
        for result in reversed(matched):
            yield result

    def _seq_event_parser(self, data_file):
        return SequentialEventReader(open(data_file.path, 'rb'), EventParser())

    def _matches(self, event, flags, match):
        if flags is not None and event.tags:
            for flag in flags:
                if not flag in event.tags:
                    return False

        if match and event.content and match:
            if not re.search(match, event.content):
                return False

        return True

    def get(self, event_id):
        """:class:`NaiveEventStore` does not support indexing, so search by ``id`` is also not supported.
        """
        log.warning('Get event by id [%s], but operation "get" is not supported in NaiveEventStore.', event_id)
        raise Exception('Get event not supported.')

    def delete(self, event_id):
        """:class:`NaiveEventStore` does not support indexing, so a delete by ``id`` is not supported.
        """
        log.warning('Delete event by id [%s], but operation "delete" is not supported in NaiveEventStore.', event_id)
        raise Exception('Delete event not supported.')



"""
-------------
theia.watcher
-------------

File watcher.

Watches files and directories for changes and emits the chnages as events.
"""
from os.path import getsize, isfile, isdir, split as splitpath, realpath
from time import time
from uuid import uuid4

from watchdog.events import FileSystemEventHandler

from theia.model import Event


class FileSource:
    """Represents a source of events.

    The underlying file that is being watched does not have to exist at the
    moment of creation of this :class:`FileSource`.

    :param str file_path: the path to the file to be watched.
    :param function callback: the callback handler to be executed when the file is
        changed. The callback is called with the difference, the path to the
        file and the list of tags for this source. The method signature looks
        like this:

        .. code-block:: python

            def callback(diff, path, tags):
                pass

        where:

        * ``diff``, ``str`` is the difference from the last state of the file.
          Usually this is the content of the emitted event.
        * ``path``, ``str`` is the path to the file that has changed. Usually
          this is the ``source`` property of the event.
        * ``tags``, ``list`` is the list of tags associated with this event
          source.

    :param str enc: the file encoding. If not specified, ``UTF-8`` is assumed.
    :param list tags: list of tags associated with this source.

    """
    def __init__(self, file_path, callback, enc='UTF-8', tags=None):
        self.path = file_path
        self.position = 0
        self.callback = callback
        self.enc = enc
        self.tags = tags or []
        self._setup()

    def modified(self):
        """Triggers an execution of the callbacks when the file has been
        modified.

        Loads the difference from the source file and calls the registered
        callbacks.
        """
        diff = self._get_diff()
        diff = diff.decode(self.enc)
        self.callback(diff, self.path, self.tags)

    def moved(self, dest_path):
        """Called when the source file has been moved to another location.

        :param str dest_path: the target location of the file after the move.
        """
        self.path = dest_path

    def created(self):
        """Called when the file has actually been created. Does not trigger the
        callbacks.
        """
        self.position = 0

    def removed(self):
        """Called when the file has been removed. Does not triggers the
        callbacks.
        """
        pass

    def _get_diff(self):
        with open(self.path, 'rb') as source_file:
            source_file.seek(self.position)
            diff = source_file.read()
            self.position = source_file.tell()
            return diff

    def _setup(self):
        if isfile(self.path):
            self.position = getsize(self.path)
        elif isdir(self.path):
            raise Exception('Not a file: %s' % self.path)


class DirectoryEventHandler(FileSystemEventHandler):
    """Implements :class:`watchdog.events.FileSystemEventHandler` and is used
    with the underlying :class:`watchdog.observers.Observer`.

    Reacts on events triggered by the watchdog Observer and passes down to the
    registered handlers.

    The handlers are registered when creating the instance as a constructor
    argument. They must be specified as ``dict`` whose keys (``str``) are the
    names of the events and the entries are the event handlers themselves.

    An example of creating new :class:`DirectoryEventHandler`:

    .. code-block:: python

        def on_file_moved(src_path, dest_path):
            print("File has moved", src_path, "->", dest_path)

        event_handler = DirectoryEventHandler(handlers={
            "moved": on_file_moved
        })

    The following events are supported:

    * ``moved`` - handles the move of a file to another location. The handler
        takes two arguments: the source path and the destination path. The
        method signature looks like this:

        .. code-block:: python

            def moved_handler(src_path, dest_path):
                pass

    * ``created`` - handles file creation. The handler takes one argument: the
        path of the created file.

        .. code-block:: python

            def created_handler(file_path):
                pass

    * ``modified`` - handles file modification. The handler takes one argument:
        the path of the modified file.

        .. code-block:: python

            def created_handler(file_path):
                pass

    * ``deleted`` - handles file deletion. The handler takes one argument: the
        path of the deleted file.

        .. code-block:: python

            def created_handler(file_path):
                pass

    :param dict handlers: a ``dict`` of handlers for specific events.

    """
    def __init__(self, handlers):
        self.handlers = handlers

    def _notify(self, event, *args):
        hnd = self.handlers.get(event)
        if hnd:
            hnd(*args)

    def on_moved(self, event):
        self._notify('moved', event.src_path, event.dest_path)

    def on_created(self, event):
        self._notify('created', event.src_path)

    def on_deleted(self, event):
        self._notify('deleted', event.src_path)

    def on_modified(self, event):
        self._notify('modified', event.src_path)


class SourcesDaemon:
    """Daemon that watches multiple sources for events.

    Uses :mod:`watchdog` to monitor files and directories for changes. This
    defaults to using ``inotify`` kernel subsystem on Linux systems, ``kqueue``
    on MacOSX and BSD-like systems and ``ReadDirectoryChangesW`` on Windows.

    :param watchdog.observers.Observer observer: an instance of the
        :class:`watchdog.observers.Observer` to be used.
    :param theia.comm.Client client: a client to a theia collector
        server.
    :param list tags: initial list of default tags that are appended to every
        file source watched by this daemon.

    """
    def __init__(self, observer, client, tags=None):
        self.sources = {}
        self.observer = observer
        self.client = client
        self.tags = tags or []
        self.dir_handler = DirectoryEventHandler(handlers={
            'moved': self._moved,
            'created': self._created,
            'deleted': self._deleted,
            'modified': self._modified
        })
        self.observer.start()

    def add_source(self, fpath, enc='UTF-8', tags=None):
        """Add source of events to be watched by this daemon.

        The path will be added as a file source and a list of tags will be
        associated with it. The default list of tags will be added to provided
        tags.

        :param str fpath: the path of the file to be watched.
        :param str enc: the file encoding. By default ``UTF-8`` is assumed.
        :param list tags: list of tags to be added to the events generated
            by this file source.

        """
        def callback(diff, _, evtags):
            """Handle an event from a :class:`FileSource` and emit an event
                to the collector server.
            """
            timestamp = time()
            event = Event(id=str(uuid4()), source=fpath, timestamp=timestamp,
                          tags=evtags, content=diff)
            self.client.send_event(event)

        fsrc = FileSource(fpath, callback, enc, self._merge_tags(tags))
        pdir, fname = self._split_path(fpath)
        files = self.sources.get(pdir)
        if not files:
            files = self.sources[pdir] = {}
            self.observer.schedule(self.dir_handler, pdir, recursive=False)
        if files.get(fname):
            return
        files[fname] = fsrc

    def _merge_tags(self, tags):
        if self.tags is None and tags is None:
            return None
        return (self.tags or []) + (tags or [])

    def remove_source(self, fpath):
        """Remove this path from the list of file event sources.

        All associated watchers and handlers are removed as well.

        :param str fpath: the path of the file to be removed from the watching
            list.

        """
        pdir, fname = self._split_path(fpath)
        files = self.sources.get(pdir)
        if files and files.get(fname):
            del files[fname]
        if self.sources.get(pdir) is not None and not self.sources[pdir]:
            del self.sources[pdir]

    @staticmethod
    def _split_path(fpath):
        fpath = realpath(fpath)
        return splitpath(fpath)

    def _get_file_source(self, src_path):
        pdir, fname = self._split_path(src_path)
        files = self.sources.get(pdir)
        if files:
            return files.get(fname)
        return None

    def _moved(self, src_path, dest_path):
        pdir, fname = self._split_path(src_path)
        files = self.sources.get(pdir)
        if files and files.get(fname):
            fsrc = files[fname]
            del files[fname]
            ndir, nfname = self._split_path(dest_path)
            if not self.sources.get(ndir):
                self.sources[ndir] = {}
            self.sources[ndir][nfname] = fsrc
            fsrc.moved(dest_path)

    def _created(self, src_path):
        fsrc = self._get_file_source(src_path)
        if fsrc:
            fsrc.created()

    def _deleted(self, src_path):
        fsrc = self._get_file_source(src_path)
        if fsrc:
            fsrc.removed()

    def _modified(self, src_path):
        fsrc = self._get_file_source(src_path)
        if fsrc:
            fsrc.modified()



"""
---------------
theia.cli.query
---------------

Theia query command line interface.
"""
import asyncio
import signal
from io import BytesIO
from datetime import datetime

from theia.query import Query
from theia.model import EventParser



def get_parser(subparsers):
    """Configures the subparser for the ``query`` command.

    :param argparse.ArgumentParser subparser: subparser for commands.

    Returns :class:`argparse.ArgumentParser` confured for the ``query`` command.
    """
    parser = subparsers.add_parser('query', help='Query for events')

    parser.add_argument('-l', '--live', action='store_true', dest='live',
                        help='Filter events in real time (live).')

    parser.add_argument('--id', dest='f_id', metavar='PATTERN',
                        default=None, help='Filter by event id')
    parser.add_argument('-s', '--source', dest='f_source', metavar='PATTERN',
                        default=None, help='Filter by event source')
    parser.add_argument('-a', '--after', dest='f_after', metavar='TIMESTAMP',
                        default=0, type=int, help='Match events after this timestamp')
    parser.add_argument('-b', '--before', dest='f_before', metavar='TIMESTAMP',
                        default=0, type=int, help='Match events before this timestamp')
    parser.add_argument('-c', '--content', dest='f_content', metavar='PATTERN',
                        default=None, help='Match event content')
    parser.add_argument('-t', '--tags', nargs='*', dest='f_tags',
                        metavar='PATTERN', help='Match any of the tags')
    parser.add_argument('-o', '--order', dest='f_order', default='asc', metavar='ORDERING',
                        help='Order of results (asc or desc). Valid only for "find".')
    parser.add_argument('-F', '--format-output', dest='o_format',
                        default='{timestamp:15} [{source:10}] {tags:15}: {content}',
                        metavar='FORMAT_STRING', help='Event output format string.' +
                        'Available properties are: id, tags, source, timestamp and content.')
    parser.add_argument('-T', '--format-timestamp', dest='o_ts_format',
                        default='%Y-%m-%d %H:%M:%S.%f%Z', metavar='DATE_FORMAT_STRING',
                        help='Timestamp strftime compatible format string')
    parser.add_argument('--close-timeout', dest='close_timeout', default=10, type=int,
                        help='Time  (in milliseconds) given to the connecting websocket to ' +
                        'actually close the connection after receiving close frame form the server. ' +
                        'The default is 10ms.')
    return parser


def format_event(event, fmt, datefmt=None):
    """Format the received event using the provided format.

    :param theia.model.Event event: the event to format.
    :param str fmt: the format string. This is compatibile with :func:`str.format`.
    :param str datefmt: alternative date format for formatiig the event timestamp.
        The format must be compatible with :func:`datetime.strftime`

    Returns the formatted event as string.
    """
    data = {
        "id": event.id,
        "tags": ",".join(event.tags),
        "source": event.source,
        "content": event.content
    }
    if datefmt:
        try:
            event_time = datetime.fromtimestamp(event.timestamp)
            data['timestamp'] = event_time.strftime(datefmt)
        except ValueError:
            data['timestamp'] = event.timestamp
    else:
        data["timestamp"] = event.timestamp

    return fmt.format(**data)


_CRITERIA_MAP = {
    'source': 'f_source',
    'id': 'f_id',
    'start': 'f_after',
    'end': 'f_before',
    'content': 'f_content',
    'tags': 'f_tags'
}


def _to_criteria(args):
    """Builds a criteria ``dict`` based on the parser arguments.
    """
    criteria = {}
    for cr_field, arg_field in _CRITERIA_MAP.items():
        if hasattr(args, arg_field) and getattr(args, arg_field):
            criteria[cr_field] = getattr(args, arg_field)
    if not args.live:
        criteria['order'] = args.f_order
    return criteria


def event_printer(event_format, time_format, parser):
    """Builds an event printer callback with the given event format and alternative
    date-time format.

    :param str event_format: the event format string.
    :param str time_format: alternative format for the event timestamp.
    :param theia.model.EventParser: event parser

    Returns event printer that takes an event data and formats is based on the
    above formats.
    """

    def print_event(evdata):
        """Formats and prints the event data.

        :param evdata: the received event data.
        """
        if isinstance(evdata, str):
            print(evdata)
            return
        try:
            event = parser.parse_event(BytesIO(evdata))
            print(format_event(event, event_format, time_format))
        except Exception as e:
            print('> Failed to show event', e)
            try:
                print('Value received: ', evdata.decode('UTF-8'))
            except:
                print('Raw data:', evdata)

    return print_event


def run_query(args):
    """Configures and runs a :class:`theia.query.Query`.

    :param argparse.Namespace args: parsed command-line arguments.
    """
    loop = asyncio.get_event_loop()
    query = Query(host=args.server_host, port=args.port, loop=loop)

    criteria_filter = _to_criteria(args)

    result = None
    parser = EventParser('UTF-8')

    if args.live:
        result = query.live(criteria_filter, event_printer(args.o_format,
                                                           args.o_ts_format,
                                                           parser))
    else:
        result = query.find(criteria_filter, event_printer(args.o_format,
                                                           args.o_ts_format,
                                                           parser))

    def clean_stop(*arg):
        """Perform a clean stop - wait a bit for the tasks on the loop to finish.
        """
        # delay the stop a bit, give it chance to actually close the loop
        loop.call_later(args.close_timeout/1000, loop.stop)

    result.when_closed(clean_stop)

    loop.add_signal_handler(signal.SIGHUP, loop.stop)
    loop.add_signal_handler(signal.SIGINT, loop.stop)
    loop.add_signal_handler(signal.SIGTERM, loop.stop)
    loop.run_forever()
    loop.close()



from theia.cli.parser import get_parent_parser as parent_parser


def get_parser(subparsers):
    parser = subparsers.add_parser('watch', help='Watcher daemon')

    parser.add_argument('-c', '--collector-server', help='Collector server name (or IP)')
    parser.add_argument('-p', '--collector-port', default=6433, help='Collector port')
    parser.add_argument('--secure', action='store_true', help='Secure connection to collector')

    parser.add_argument('-f', '--files', nargs='*', dest='files',
                        help='List of files to watch for changes')
    parser.add_argument('-d', '--directory', nargs='*', dest='dirs',
                        help='List of directories to whose files are watched for changes')

    parser.add_argument('-t', '--tags', nargs='*', dest='tags', help='List of tags')

    return parser


def run_watcher(args):
    import socket
    import asyncio
    import functools
    import signal
    from theia.comm import Server, Client
    from watchdog.observers import Observer
    from theia.watcher import SourcesDaemon

    from threading import Thread

    hostname = socket.gethostname()
    print(hostname)

    loop = asyncio.get_event_loop()

    client = Client(loop=loop, host=args.collector_server,
                    port=args.collector_port, path='/event', secure=args.secure)

    client.connect()

    daemon = SourcesDaemon(observer=Observer(), client=client, tags=[hostname])

    for f in args.files:
        daemon.add_source(fpath=f, tags=args.tags)

    loop.add_signal_handler(signal.SIGHUP, loop.stop)
    loop.add_signal_handler(signal.SIGINT, loop.stop)
    loop.add_signal_handler(signal.SIGTERM, loop.stop)
    loop.run_forever()
    loop.close()
    print("Watcher stopped")



"""
-------------------
theia.cli.collector
-------------------

Theia collector command line interface script.
"""
import signal
from logging import getLogger
from theia.naivestore import NaiveEventStore
from theia.collector import Collector


log = getLogger(__name__)


def get_parser(subparsers):
    """Configures the subparser for the ``collect`` command.

    :param argparse.ArgumentParser subparser: subparser for commands.

    Returns :class:`argparse.ArgumentParser` confured for the ``collect`` command.
    """
    parser = subparsers.add_parser('collect', help='Collector server')

    parser.add_argument('-d', '--data-dir', dest='data_dir', help='Data store root directory')
    parser.add_argument('-U', '--db-url', dest='db_url', help='Database URL (SQLAlchemy form)',
                        default=None)
    parser.add_argument('--verbose', dest='store_verbose', action='store_true',
                        help='Make the EventStore to log more verbose output.')
    parser.add_argument('--rdbs-store', dest='rdbs_store', action='store_true',
                        help='Use RDBS EventStore instead of NaiveEventStore.' +
                        'The RDBS store keeps the events in a relational database.')

    return parser


def run_collector(args):
    """Runs the collector server.

    :param argparse.Namespace args: arguments to configure the
        :class:`theia.collector.Collector` instance.

    """
    store = None

    if args.rdbs_store:
        store = get_rdbs_store(args)
        if store:
            log.info('Using RDBS Event Store')
        else:
            log.warning('Unable to set up the RDBS Event Store. Will fall back to using Naive Event Store.')
    if not store:
        store = get_naive_store(args)
        log.info('Using Naive Event Store')

    collector = Collector(store=store, hostname=args.server_host, port=args.port)

    def stop_collector(sig, frame):
        """Signal handler that stops the collector.
        """
        log.info('Collector is shutting down.')
        collector.stop()

    signal.signal(signal.SIGHUP, stop_collector)
    signal.signal(signal.SIGINT, stop_collector)
    signal.signal(signal.SIGTERM, stop_collector)

    collector.run()


def _has_sqlalchemy():
    try:
        import sqlalchemy
    except:
        return False
    return True


def get_rdbs_store(args):
    """Creates and configures new :class:`theia.rdbs.RDBSEventStore` based on
    the arguments passed.

    :param argparse.Namespace args: arguments.

    """
    if not _has_sqlalchemy():
        log.info('SQLAlchemy is not present on your system. RDBSEventStore cannot work without it.')
        return None

    from theia.rdbs import create_store

    if not args.db_url:
        raise Exception('No Database URL')

    return create_store(db_url=args.db_url, verbose=args.store_verbose)


def get_naive_store(args):
    """Creates and configures new :class:`theia.naivestore.NaiveEventStore`
    based on the arguments passed.

    :param argparse.Namespace args: arguments.

    """

    if not args.data_dir:
        raise Exception('No data directory specified')

    return NaiveEventStore(root_dir=args.data_dir)



"""
----------
theia.rdbs
----------

Relational database EventStore implementation.
"""
import re
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Float, create_engine, desc
from sqlalchemy.orm import sessionmaker
from theia.storeapi import (EventStore,
                            EventWriteException,
                            EventReadException,
                            EventNotFound,
                            EventStoreException)
from theia.model import Event


Base = declarative_base()


class EventRecord(Base):
    """SQLAlchemy model representing the event.
    """

    __tablename__ = 'events'

    id = Column(String, primary_key=True)
    timestamp = Column(Float, index=True)
    tags = Column(String)
    source = Column(String, index=True)
    content = Column(String, index=True)
    custom_headers = Column(String)

    def __repr__(self):
        return 'Event<%s @ %s>' % (self.id, str(self.timestamp))

    def __str__(self):
        return self.__repr__()

    def __eq__(self, obj):
        if obj is None:
            return False
        if not isinstance(obj, EventRecord):
            return False
        return self.id == obj.id

    def __hash__(self):
        return hash(self.id)


class RDBSEventStore(EventStore):
    """EventStore that persists the events in a relational database.

    The implementation relies on SQLAlchemy ORM framework.

    :param session_factory: the SQLAlchemy SessionMaker function.
    """

    def __init__(self, session_factory):
        self.session_factory = session_factory
        self.bulk_size = 128

    def _session(self):
        """Creates new session.
        """
        return self.session_factory()

    def save(self, event):
        """Stores the event in the underlying database.

        Note that this method only does *INSERT* as the ``EventStore`` has no concept of *UPDATE* -
        each event is only added and cannot be updated. Adding the same event twice will result in
        an error.

        :param event (theia.model.Event): the event to save.

        This method does not return a value.
        """
        event_record = EventRecord(id=event.id,
                                   timestamp=float(event.timestamp),
                                   tags=','.join((event.tags or [])),
                                   source=event.source,
                                   content=event.content)
        sess = self._session()
        try:
            sess.add(event_record)
            sess.commit()
        finally:
            sess.close()

    def delete(self, event_id):
        """Deletes the event with the given event id.

        :param event_id(str): the event id
        """
        sess = self._session()
        try:
            event_record = sess.query(EventRecord).get(event_id)
            if not event_record:
                raise EventNotFound()
            sess.delete(event_record)
            sess.commit()
        except Exception as e:
            raise EventWriteException(str(e)) from e
        finally:
            sess.close()

    def get(self, event_id):
        """Looks up an event by its id.

        :param event_id(str): the event id.

        Returns a ``theia.model.Event`` if the event was found or ```theia.storeapi.EventNotFound``` if
        no such event can be found.
        """
        sess = self._session()
        try:
            event_record = sess.query(EventRecord).get(event_id)
            if not event_record:
                raise EventNotFound()
            return Event(id=event_record.id,
                         timestamp=str(event_record.timestamp),
                         tags=(event_record.tags or '').split(','),
                         source=event_record.source,
                         content=event_record.content)
        except EventNotFound as ne:
            raise ne
        except Exception as e:
            raise EventReadException(str(e)) from e
        finally:
            sess.close()

    def search(self, ts_start, ts_end=None, flags=None, match=None, order='asc'):
        """Performs a search through the stored events.

        :param ts_start(float): *required*, match all events that occured at or later than this time.
        :param ts_end(float): *optional*, match all events that occured before this time.
        :param flags(list): *optional*, list of string values (regular expressions) against which to match
        the event tags. Event matches only if *all* flags are matched against the event's tags.
        :param match(str): *optional*, match the content of an event. This is a regular expession as well.
        :param order(str): either ``asc`` or ``desc`` - sort the results ascending or descending based on the
        event timestamp.

        Returns an iterator over all matched results.
        """
        if not ts_start:
            raise EventStoreException('start timestamp is required when searching events')

        flag_matchers = None
        content_matcher = None
        if flags:
            try:
                flag_matchers = [re.compile(f) for f in flags]
            except Exception:
                raise EventStoreException('invalid flags match regex.')

        if match:
            try:
                content_matcher = re.compile(match)
            except Exception:
                raise EventStoreException('invalid content match regex.')

        def get_bulk(blk):
            """Fetch single page bulk results.
            """
            sess = self._session()
            qry = sess.query(EventRecord).filter(EventRecord.timestamp >= ts_start)

            if ts_end:
                qry.filter(EventRecord.timestamp <= ts_end)
            if order == 'asc':
                qry.order_by(EventRecord.timestamp)
            else:
                qry.order_by(desc(EventRecord.timestamp))

            results = []
            count = 0
            for event_record in qry.limit(self.bulk_size).offset(blk*self.bulk_size).all():
                count += 1
                flags = (event_record.tags or '').split(',')

                if flag_matchers:
                    if not match_all(flag_matchers, flags):
                        continue

                if content_matcher:
                    if not content_matcher.fullmatch(event_record.content or ''):
                        continue

                event = Event(id=event_record.id,
                              timestamp=str(event_record.timestamp),
                              tags=flags,
                              source=event_record.source or '',
                              content=event_record.content)

                results.append(event)
            return results, blk, count > 0

        page = 0
        has_more = True
        while has_more:
            results, page, has_more = get_bulk(page)
            for result in results:
                yield result
            page += 1

    def close(self):
        """Closes the store.

        Does nothing in this implementation.
        """
        pass

def match_any(matcher, values):
    """Check if the matcher matches *any* of the supplied values.

    :param matcher(regex pattern): compiled regular expression.
    :param values(list): list of ``str`` values to match.

    Returns ``True`` if *any* of the ``str`` values matches (fullmatch) the matcher;
    otherwise ``False``.
    """
    for val in values:
        if matcher.fullmatch(val):
            return True
    return False


def match_all(matchers, values):
    """Check if *all* matchers matchy any of the given values.

    Each matcher *must* match at least one value of the list of values.

    :param matchers(list): list of compiled regular expressions.
    :param values(list): list of ``str`` to match

    Returns ``True`` only if *all* of the matchers have matched at least one value of the provided
    list of values.
    """
    for matcher in matchers:
        if not match_any(matcher, values):
            return False
    return True

def create_store(db_url, verbose=False):
    """Creates new RDBSEventStore.

    :param db_url(str): The database URL in SQLAlchemy form.
    :param verbose(bool): ``True`` to show extended messages from the store.

    Returns RDBSEventStore object.
    """
    engine = create_engine(db_url, echo=verbose)
    Base.metadata.create_all(engine)
    session_factory = sessionmaker(bind=engine)

    return RDBSEventStore(session_factory)



"""
-----------
theia.model
-----------

Theia event data model.

Basic model of an Event, serializers and parsers for Event manipulation.
"""

from time import time
from collections import namedtuple
from io import StringIO, SEEK_CUR
import re


EventPreamble = namedtuple('EventPreamble', ['total', 'header', 'content'])
"""A preamble to an :class:`Event`.

The preamble is present only in the serialized representation of an event and
holds the information about the size of the event and its parts.
"""

EventPreamble.total.__doc__ = """
    ``int``, the size of the serialized event in bytes.
"""

EventPreamble.header.__doc__ = """
    ``int``, the size of the serialized event header in bytes.
"""

EventPreamble.content.__doc__ = """
    ``int``, the size of the serialized event content in bytes.
"""


class Header:
    """Header represents an Event header. The header contains the following
    properties:

    * ``id``, unique identifier for the event. Usually UUIDv4.
    * ``timestamp``, floating point of the number of milliseconds since epoch
      start (1970-1-1T00:00:00.00).
    * ``source``, string, the name of the event source.
    * ``tags``, list of strings, arbitrary tags attached to the event.

    The header is usefull and usually used when serializing/parsing an event.
    """

    def __init__(self, id=None, timestamp=None, source=None, tags=None):
        self.id = id
        self.timestamp = timestamp
        self.source = source
        self.tags = tags


class Event:
    """Event represnts some event occuring at a specfic time.

    Each event is uniquely identified by its ``id`` in the whole system. An event
    comes from a ``source`` and always has an associated ``timestamp``. The
    timestamp is usually generated by the event producer.

    The *content* of an event is an arbitrary string. It may be a log file line,
    some generated record, readings from a sensor or other non-structured or
    structured text.

    Each event may have a list of ``tags`` associcated with it. These are arbitrary
    strings and help in filtering the events.

    An event may look like this ::

        id:331c531d-6eb4-4fb5-84d3-ea6937b01fdd
        timestamp: 1509989630.6749051
        source:/dev/sensors/door1-sensor
        tags:sensors,home,doors,door1
        Door has been unlocked.

    The constructor takes multiple arguments, of which only the id and source are
    required.

    :param id: ``str``, the event unique identifier. Must be system-wide unique. An
        UUID version 4 (random UUID) would be a good choice for ``id``.
    :param source: ``str``, the source of the event. It usually is the name of the
        monitored file, but if the event does not originate from a file, it should be
        set to the name of the process, system or entity that generated the event.
    :param timestamp: ``float``, time when the event occured in seconds (like UNIX
        time). The value is a floating point number with nanoseconds precission. If
        no value is given, then the current time will be used.
    :param tags: ``list``, list of ``str`` tags to add to this event.
    :param content: ``str``, the actual content of the event. The content may have
        an arbitrary lenght (or none at all).

    """

    def __init__(self, id, source, timestamp=None, tags=None, content=None):
        self.id = id
        self.source = source
        self.timestamp = timestamp or time()  # time in nanoseconds UTC
        self.tags = tags or []
        self.content = content or ''

    def match(self, id=None, source=None, start=None, end=None, content=None, tags=None):
        """Check if this event matches the provided criteria.

        The event will match only if **all** criteria is statisfied. Calling match
        without any criteria, yields ``True``.

        The criteria is processed as a regular expression. Each value is first
        converted to string, then matched against the provided regular expression -
        see :func:`re.match`. The exception of this rule are the criteria for
        ``start`` and ``end`` wich expect numeric values, as they operate on the
        :class:`Event` timestamp.

        :param id: ``str``, regular expression against which to match the :class:`Event`
            ``id``.
        :param source: ``str``, regular expression against which to match the :class:`Event`
            ``source``.
        :param start: ``float`` or ``int``, match true if the :class:`Event` timestamp
            is greater than or equal to this value.
        :param start: ``float`` or ``int``, match true if the :class:`Event` timestamp
            is less than or equal to this value.
        :param content: ``str``, regular expression against which to match the :class:`Event`
            ``content``.
        :param tags: ``list``, list of ``str`` regular expressions against which to
            match the :class:`Event` tags. Matches true only if **all** of the provided
            criteria match the Event tags.

        Returns ``True`` if this :class:`Event` matches the criteria, otherwise ``False``.
        """
        return all([self._match_header_id_and_source(id, source),
                    self._match_timestamp(start, end),
                    self._match_tags(tags),
                    self._match_content(content)])

    def _match_header_id_and_source(self, id, source):
        matches = True
        if id is not None:
            matches = _match(id, self.id)
        if matches and source is not None:
            matches = _match(source, self.source)
        return matches

    def _match_timestamp(self, start, end):
        matches = True
        if self.timestamp:
            if start is not None:
                matches = self.timestamp >= start
            if matches and end is not None:
                matches = self.timestamp <= end
        return matches

    def _match_tags(self, tags):
        if tags:
            for tag in tags:
                if tag not in self.tags:
                    return False
        return True

    def _match_content(self, content):
        if not content:
            return True
        return _match(content, self.content)


def _match(pattern, value):
    """Match the value against a regular expression pattern.
    """
    if value is None:
        return False
    return re.match(pattern, value) is not None


class EventSerializer:
    """Serialized for instances of type :class:`Event`.

    This serializes the :class:`Event` in a plain text representation of the Event.
    The serialized text is encoded in UTF-8 and the actual ``bytes`` are returned.

    The representation consists of three parts: preamble, header and content.
    The preamble is the first line of every event and has this format:

        event: <total_size> <header_size> <content_size>

    where:

    * ``total_size`` is the total size of the Event (after the heading) in bytes.
    * ``header_size`` is the size of the header in bytes.
    * ``content_size`` is the size of the content (after the Header) in bytes.

    The header holds the values for the Event's id, source, tags and timestamp.
    Each value is serialized on a single line. The line starts with the name of
    the property, separated by a colon(``:``), then the property value.

    The content starts after the final header and is separated by a newline.

    Here is an example of a fully serialized :class:`Event` (Python's ``bytes``)::

        b'event: 155 133 22\\nid:331c531d-6eb4-4fb5-84d3-ea6937b01fdd\\ntimestamp: 1509989630.6749051\\nsource:/dev/sensors/door1-sensor\\ntags:sensors,home,doors,door1\\nDoor has been unlocked\\n'

    or as a textual representation::

        event: 155 133 22
        id:331c531d-6eb4-4fb5-84d3-ea6937b01fdd
        timestamp: 1509989630.6749051
        source:/dev/sensors/door1-sensor
        tags:sensors,home,doors,door1
        Door has been unlocked

    **Note** that the :class:`EventSerializer` adds a trailing newline (``\\n``) at
    the end.

    The serializer constructor takes the encoding as an argument. By default "utf-8"
    is used.

    :param encoding: ``str``, the encoding of the serialized string. Default ``utf-8``.

    """
    def __init__(self, encoding='utf-8'):
        self.encoding = encoding

    def serialize(self, event):
        """Serializes an :class:`Event`.

        See :class:`EventSerializer` for details on the serialization format.

        :param event: :class:`Event`, the event to be serialized.

        Returns the serialized event as ``bytes``.
        """
        event_str = ''
        hdr = self._serialize_header(event)
        hdr_size = len(hdr.encode(self.encoding))
        cnt = event.content or ''
        cnt_size = len(cnt.encode(self.encoding))
        total_size = hdr_size + cnt_size
        event_str += 'event: %d %d %d\n' % (total_size, hdr_size, cnt_size)
        event_str += hdr
        event_str += cnt
        event_str += '\n'
        return event_str.encode(self.encoding)

    def _serialize_header(self, event):
        hdr = ''
        hdr += 'id:' + str(event.id) + '\n'
        hdr += 'timestamp: %.7f' % event.timestamp + '\n'
        hdr += 'source:' + str(event.source) + '\n'
        hdr += 'tags:' + ','.join(event.tags) + '\n'
        return hdr


_HEADER_FIELDS = {
    'id': lambda value: value,
    'timestamp': lambda value: float(value),
    'source': lambda value: value,
    'tags': lambda value: value.split(',')
}


class EventParser:
    """Parses an incoming bytes stream into an :class:`Event`.

    Offers methods for parsing parts of an :class:`Event` or parsing the full
    event from the incoming :class:`io.BytesIO` stream.

    The stream will be decoded before converting it to ``str``. By default the
    parser assumes that the stream is ``utf-8`` encoded.

    :param encoding: ``str``, the encoding to be ued for decoding the stream bytes.
        The default is ``utf-8``.

    """
    def __init__(self, encoding='utf-8'):
        self.encoding = encoding

    def parse_header(self, hdr_size, stream):
        """Parses the :class:`Event` header into a :class:`Header` object from the
        incoming stream.

        First ``hdr_size`` bytes are read from the :class:`io.BytesIO` stream and
        are decoded to ``str``.

        Then, the parser parses each line by splitting it by the first colon (``:``).
        The first part is ued to determine the :class:`Header` property. The part
        after the colon is the propery value.

        :param hdr_size: ``int``, the size of the header in bytes. See
            :meth:`EventParser.parse_preamble` on how to determine the header size
            in bytes.
        :param stream: :class:`io.BytesIO`, the incoming stream to parse.

        Returns the parsed :class:`Header` for the event.

        Raises :class:`Exception` if an unknown property is encountered in the header.
        """
        hbytes = stream.read(hdr_size)
        if len(hbytes) != hdr_size:
            raise Exception('Invalid read size from buffer. The stream is either unreadable \
                             or corrupted. %d read, expected %d' % (len(hbytes), hdr_size))
        hdr_str = hbytes.decode(self.encoding)
        header = Header()
        sio = StringIO(hdr_str)

        line = sio.readline()
        while line:
            line = line.strip()
            if not line:
                raise Exception('Invalid header')
            idx = line.index(':')
            prop = line[0:idx].lower()
            value = line[idx + 1:]
            parse_field = _HEADER_FIELDS.get(prop)
            if not parse_field:
                raise Exception('Unknown property in header %s' % prop)
            setattr(header, prop, parse_field(value))
            line = sio.readline()
        sio.close()
        return header

    def parse_preamble(self, stream):
        """Parses the event preamble from the incoming stream into a :class:`EventPreamble`.

        The event preamble is a single line read from the stream with the following
        structure::

            <total_size> <header_size> <content_size>

        where:

        * ``total_size`` is the total size of the Event (after the heading) in bytes.
        * ``header_size`` is the size of the header in bytes.
        * ``content_size`` is the size of the content (after the Header) in bytes.

        Note that the sizes are expressed in bytes.

        :param stream: :class:`io.BytesIO`, the stream to parse.

        Returns the parsed :class:`EventPreamble` from the incoming stream.
        """
        pstr = stream.readline()
        if not pstr:
            raise EOFException()
        if pstr:
            pstr = pstr.decode(self.encoding).strip()

        if not pstr.startswith('event:'):
            raise Exception('Invalid preamble line: [%s]' % pstr)

        values = pstr[len('event:') + 1:].split(' ')
        if len(values) != 3:
            raise Exception('Invalid preamble values')

        return EventPreamble(total=int(values[0]), header=int(values[1]), content=int(values[2]))

    def parse_event(self, stream, skip_content=False):
        """Parses an event from the incoming stream.

        The parsing is done in two phases:

        1. The preamble is parsed to determine the total size of the event and the
            the size of the event header.
        2. Then the actual event is read, either with or without the event content.

        If ``skip_content`` is set to ``True``, then the actual content of the event
        is not read. This is usefull in event readers that do matching of the event
        header values, without wasting memory and performance for reading the content.
        In this case, the event content will be set to ``None``.

        :param stream: :class:`io.BytesIO`, the stream to parse from.
        :param skip_content: ``bool``, whether to skip the fetching of the content and
            to fetch only the :class:`Event` header. Default is ``False``.

        Returns the parsed :class:`Event` from the incoming stream.
        """
        preamble = self.parse_preamble(stream)
        header = self.parse_header(preamble.header, stream)
        content = None
        read_length = None
        if skip_content:
            stream.seek(preamble.content, SEEK_CUR)
        else:
            content = stream.read(preamble.content)
            read_length = len(content)
            content = content.decode(self.encoding)

        stream.seek(1, SEEK_CUR)  # new line after each event

        if not skip_content and read_length != preamble.content:
            raise Exception('Invalid content size. The stream is either unreadable or corrupted. ' +
                            'Preamble declares %d bytes, but content length is %d' % (preamble.content, len(content)))

        return Event(id=header.id, source=header.source, timestamp=header.timestamp,
                     tags=header.tags, content=content)


class EOFException(Exception):
    """Represents an error in parsing an :class:`Event` from an underlyign stream
    that occurs when the end of stream is reached prematurely.
    """
    pass



from theia.comm import Client
from theia.model import Event, EventSerializer

from random import randint
from uuid import uuid4
from datetime import datetime
from threading import Thread

import argparse
import lorem
import asyncio
import time
import random


def get_parser():
    parser = argparse.ArgumentParser(prog='theia.cli.simulate',
                                     description='Simulate events (debugging)')

    parser.add_argument('-H', '--host', default='localhost', dest='host', help='Collector host')
    parser.add_argument('-p', '--port', default=6433, dest='port', help='Collector port')
    parser.add_argument('-t', nargs='*', dest='tags', help='Set of tags to choose from')
    parser.add_argument('-s', nargs='*', dest='sources', help='Set of event sources to choose from')
    parser.add_argument('-c', dest='content', default=None,
                        help='Use this event content instead of random content.')
    parser.add_argument('--content-size', dest='content_size', type=int,
                        default=None, help='Size of content (approximately)')
    parser.add_argument('--delay', dest='delay', default=1.0, type=float,
                        help='Delay between event in seconds')

    return parser


def generate_content(content=None, size=None):
    if content is not None:
        return content

    if size is None:
        return lorem.sentence()

    content = lorem.sentence()

    while len(content) < size:
        content = content + ' ' + lorem.sentence()

    return content


def generate_rand_items(items, default):
    if len(items) == 0:
        return [default]

    if len(items) == 1:
        return items

    rn = randint(1, len(items))

    rndi = [n for n in items]
    random.shuffle(rndi)

    return rndi[0:rn]


def generate_rand_event(sources, tags, content, cnt_size):
    source = sources[randint(0, len(sources) - 1)]
    e = Event(id=str(uuid4()), source=source, timestamp=datetime.now().timestamp(),
              tags=generate_rand_items(tags, 'tag-4'), content=generate_content(content, cnt_size))
    return e


def simulate_events(args):
    loop = asyncio.get_event_loop()
    client = Client(host=args.host, port=args.port, path='/event', loop=loop)

    print('tags', args.tags)
    client.connect()

    ser = EventSerializer()
    tags = [args.tags] if isinstance(args.tags, str) else args.tags

    def send_event():
        e = generate_rand_event(args.sources, tags, args.content, args.content_size)
        client.send_event(e)
        print(ser.serialize(e).decode('UTF-8'))
        print()

    def send_all():
        while True:
            send_event()
            time.sleep(args.delay)

    t = Thread(target=send_all)
    t.start()
    loop.run_forever()


if __name__ == '__main__':
    parser = get_parser()

    args = parser.parse_args()

    simulate_events(args)



"""
--------------
theia.storeapi
--------------

Event Store API
^^^^^^^^^^^^^^^

Defines classes, methods and exceptions to be used when implementing an Event Store.
"""
from abc import abstractmethod


class EventStore:
    """EventStore is the basic interface for interaction with the events.

    Main uses of this store are CRUD interactions with the events. The API
    provides powerful search through all events based on a time range and
    optionally additional flags.
    An instance of this class is thread-safe.
    """
    @abstractmethod
    def save(self, event):
        """Saves an event in the underlying storage.

        This method is guaranteed to be atomic in the sense that the storage will
        either succeed to write and flush the event, or it will fail completely. In
        either case, the storage will be left in a consistent state.

        :param event: :class:`theia.model.Event`, the Event object to store.

        This method does not return any value.
        """
        pass

    @abstractmethod
    def delete(self, event_id):
        """Deletes an event from the storage.

        The delete operation removes an event from the underlying storage. This
        operation is guaranteed to be atomic, the event will either be removed or
        it will fail completely. In either case the storage will be left in a
        consistent state.

        :param event_id: ``str``, the unique identifier of the event to be removed.

        This method does not return any value.
        """
        pass

    @abstractmethod
    def get(self, event_id):
        """Looks up an event by its unique identifier.

        The storage will try to look up the event with the specified id:

        * if the event is found, it will return an Event object
        * the event is not found, raises an EventNotFound excepton.

        Edge cases:

        * If the event is being inserted AFTER the get(..) operation is invoked,
            there is NO guarantee that it will be fetched.
        * If the event is being inserted BEFORE the get(..)  operation is invoked,
            but that transaction is still not commited, the operaton will block
            until the write operation completes (or errors out) and the Event will
            be returned (if the write succeds) or will error out (if the write
            fails) - strict consistency

        **Note:**
            Some specific implementations may break the strict consistency if the
            underlying mechanism does not provide means to implement it. In those
            cases, the subclass must override this documentation and must document
            its exact for the above edge-cases.

        :param event_id: ``str``, the unique identifier of the event to be looked up.

        Returns the :class:`theia.model.Event` with the given id, or ``None`` if no such
        event exists.
        """
        pass

    @abstractmethod
    def search(self, ts_start, ts_end=None, flags=None, match=None, order='asc'):
        """Performs a search for events matching events in the specified time range.

        :param ts_start: ``float``, start of the time range. Matching events with timestamp bigger
            or equal to this paramter will be returned.
        :param ts_end: ``float``, end of the time range. Matching events with timestamp smaller or
            equal to this paramter will be returned.
        :param flags: ``list``,  events that have ALL of the flags will be returned.
        :param match: ``str``, regular expression, (restricted to a subset of the full regexp
            support) to match the event content against.
        :param order: ``str``, ``'asc'`` or ``'desc'``, order in which the events are returned.

        The operation returns an iterator over the matched (ordered) set of events.
        This operation satisfies the strict consistency.
        """
        pass

    @abstractmethod
    def close(self):
        """Close and cleanup the underlying store.
        """
        pass


class EventStoreException(Exception):
    """General store error.
    """
    pass


class EventWriteException(EventStoreException):
    """Represents an error while writing an event to the underlying storage.
    """
    pass


class EventReadException(EventStoreException):
    """Represents an error while reading an event from the underlying storage.
    """
    pass


class EventNotFound(EventReadException):
    """Raised if there is no event found in the underlying storage.
    """
    pass



"""
----------------
theia.cli.parser
----------------


Theia CLI main :mod:`argparse` parser.
"""
import argparse


def get_parent_parser(name, desc=''):
    """Creates the main (parent) :class:`argparse.ArgumentParser` for Theia CLI.

    Defines the main argument options such as theia server host, port, verbosity
    level etc.

    :param str name: the name of the program.
    :param str desc: program description.

    Returns the configured :class:`argparse.ArgumentParser`.
    """
    parser = argparse.ArgumentParser(prog=name, description=desc)

    parser.add_argument('-v', '--version',
                        help='Print program version and exit', action='store_true')
    parser.add_argument('-H', '--host', help='Hostname to bind to',
                        default='localhost', dest='server_host')
    parser.add_argument('-P', '--port', help='Listen on port', default=6433,
                        type=int)

    parser.add_argument('--verbose', dest='verbose', action='store_true',
                        help='Verbose output.')

    return parser



"""
---------------
theia.collector
---------------

The log aggregator collector server imlementation.
"""
import asyncio
from io import BytesIO
import json
from logging import getLogger
from threading import Thread
from theia.comm import Server
from theia.model import EventParser, EventSerializer



log = getLogger(__name__)

# pylint: disable=too-few-public-methods
# This needs only one method.
class LiveFilter:
    """Filter for the live event pipe.

    Holds a single criteria to filter events by.

    :param ws: :class:`websockets.WebSocketClientProtocol`, reference to the web socket instance.
    :param criteria: ``dict``,  dict holding criteria values.
    """

    ALLOWED_CRITERIA = {'id': str, 'source': str,
                        'start': int, 'end': int, 'content': str, 'tags': list}

    def __init__(self, ws, criteria):
        self.ws = ws
        self.criteria = criteria
        self._check_criteria()

    def _check_criteria(self):
        for k, value in self.criteria.items():
            allowed = LiveFilter.ALLOWED_CRITERIA[k]
            if allowed is None:
                raise Exception('unknown criteria %s' % k)
            if not isinstance(value, allowed):
                raise Exception('invalid value for criteria %s' % k)

    def match(self, event):
        """Matches an event to the criteria of this filter.

        :param event: :class:`theia.model.Event`, the event to match.

        Returns:
            match(``bool``), True if the event matches the criteria, otherwise False.
        """

        return event.match(**self.criteria)


class Live:
    """Live event pipeline.

    Each event is passed through the live event pipeline and matched to the
    LiveFilter filters.

    :param serializer: :class:`theia.model.Serializer`, event serializer.
    """

    def __init__(self, serializer):
        self.serializer = serializer
        self.filters = {}

    def add_filter(self, lfilter):
        """Adds new filter to the pipeline.

        :param lfilter: :class:`LiveFilter`, the filter to add to the pipeline.
        """
        self.filters[lfilter.ws] = lfilter

    async def pipe(self, event):
        """Add an event to the live pipeline.

        The event will be matched against all filters in this pipeline.

        :param event: :class:`theia.model.Event`, the event to be pipelined.
        """
        for ws, live_filter in self.filters.items():
            if live_filter.match(event):
                try:
                    result = None
                    try:
                        result = self.serializer.serialize(event)
                    except Exception as se:
                        # failed to serialize
                        result = json.dumps({'error': True, 'message': str(se)})
                        log.debug('Serialization error: %s', se)
                    await ws.send(result)
                except Exception as e:
                    log.error('Something went wrong: %s', e)
                    log.exception(e)


class Collector:
    """Collector server.

    Collects the events, passes them down the live pipe filters and stores them
    in the event store.

    :param store: :class:`theia.storeapi.Store`, store instance
    :param hostame: ``str``, server hostname. Default is '0.0.0.0'.
    :param port: ``int``, server port. Default is 4300.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self, store, hostname='0.0.0.0', port=4300):
        self.hostname = hostname
        self.port = port
        self.server = None
        self.store = store
        self.server_thread = None
        self.store_thread = None
        self.store_loop = None
        self.server_loop = None
        self.parser = EventParser()
        self.serializer = EventSerializer()
        self.live = Live(self.serializer)

    def run(self):
        """Run the collector server.

        This operation is blocking.
        """
        self._setup_store()
        self._setup_server()
        self.store_thread.join()
        self.server_thread.join()

    def stop(self):
        """Stop the collector server.

        This operation is non-blocking.
        """

        self.server.stop()
        try:
            self.store_loop.call_soon_threadsafe(self.store_loop.stop)
        finally:
            self.server_loop.call_soon_threadsafe(self.server_loop.stop)
        self.store.close()

    def _setup_store(self):
        def run_store_thread():
            """Runs the store loop in a separate thread.
            """

            loop = asyncio.new_event_loop()
            self.store_loop = loop
            loop.run_forever()
            loop.close()
            log.info('store is shut down.')

        self.store_thread = Thread(target=run_store_thread)
        self.store_thread.start()

    def _setup_server(self):
        def run_in_server_thread():
            """Runs the server loop in a separate thread.
            """

            loop = asyncio.new_event_loop()
            self.server_loop = loop
            self.server = Server(loop=loop, host=self.hostname, port=self.port)
            self.server.on_action('/event', self._on_event)
            self.server.on_action('/live', self._add_live_filter)
            self.server.on_action('/find', self._find_event)
            self.server.start()
            loop.run_forever()
            loop.close()
            log.info('server is shut down.')

        self.server_thread = Thread(target=run_in_server_thread)
        self.server_thread.start()

    def _on_event(self, path, message, websocket, resp):
        try:
            self.store_loop.call_soon_threadsafe(self._store_event, message)
        except Exception as e:
            log.exception(e)

    def _store_event(self, message):
        event = self.parser.parse_event(BytesIO(message))
        self.store.save(event)
        try:
            asyncio.run_coroutine_threadsafe(self.live.pipe(event), self.server_loop)
        except Exception as e:
            log.error('Error in pipe: %s (event: %s)', e, event)

    def _add_live_filter(self, path, message, websocket, resp):
        criteria = json.loads(message)
        live_filter = LiveFilter(websocket, criteria)
        self.live.add_filter(live_filter)
        return 'ok'

    def _find_event(self, path, message, websocket, resp):
        criteria = json.loads(message)
        ts_from = criteria.get('start')
        ts_to = criteria.get('end')
        flags = criteria.get('tags')
        content = criteria.get('content')
        order = criteria.get('order') or 'asc'
        if not ts_from:
            raise Exception('Missing start timestamp')
        self.store_loop.call_soon_threadsafe(
            self._find_event_results, ts_from, ts_to, flags, content, order, websocket)
        return 'ok'

    def _find_event_results(self, start, end, flags, match, order, websocket):
        for event in self.store.search(ts_start=start, ts_end=end, flags=flags,
                                       match=match, order=order):
            ser = self.serializer.serialize(event)
            asyncio.run_coroutine_threadsafe(websocket.send(ser), self.server_loop)



import setuptools


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name="theia",
    version="0.9.2",
    author="Pavle Jonoski",
    author_email="jonoski.pavle@gmail.com",
    description="Theia is a lightweight log aggregator.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/theia-log/theia",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Internet :: Log Analysis",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Logging",
        "Topic :: Utilities",
    ),
    install_requires=[
        "picotui==0.9.1",
        "watchdog==0.8.3",
        "websockets==3.3",
    ]
)

